@@ -277,7 +277,7 @@ class Queue:
         for msgid in self.msgs:
             self.count = self.count + 1
             try:
-                M = Message(msgid, self.command_recipient).initMessage()
+                M = Message(msgid, self.command_recipient)
             except Errors.MessageError, obj:
                 self.cPrint(obj)
                 continue
@@ -426,24 +426,18 @@ class Message:
         except email.Errors.BoundaryError:
             self.msgobj = Util.msg_from_file(open(self.msgfile, 'r'))
         self.recipient = recipient
-
-    def initMessage(self, recipient = None):
-        self.return_path = parseaddr(self.msgobj.get('return-path'))[1]
-        if not recipient and not self.recipient:
+        if self.recipient is None:
             self.recipient = self.msgobj.get('x-tmda-recipient')
-        else:
-            self.recipient = recipient
-        return self
+        self.return_path = parseaddr(self.msgobj.get('return-path'))[1]
+        self.x_primary_address = parseaddr(self.msgobj.get('x-primary-address'))[1]
+        self.append_address = Util.confirm_append_address(
+            self.x_primary_address, self.return_path)
 
     def release(self):
         """Release a message from the pending queue."""
-        try:
-            self.return_path
-        except:
-            self.initMessage()
         import Cookie
         if Defaults.PENDING_RELEASE_APPEND:
-            Util.append_to_file(self.return_path,
+            Util.append_to_file(self.append_address,
                                 Defaults.PENDING_RELEASE_APPEND)
         timestamp, pid, suffix = self.msgid.split('.')
         # Remove Return-Path: to avoid duplicates.
@@ -462,28 +456,19 @@ class Message:
         del self.msgobj['X-TMDA-Released']
         self.msgobj['X-TMDA-Released'] = Util.unixdate()
         # Reinject the message to the original envelope recipient.
-        Util.sendmail(Util.msg_as_string(self.msgobj),
-                      self.recipient, self.return_path)
+        Util.sendmail(self.show(), self.recipient, self.return_path)
 
     def delete(self):
         """Delete a message from the pending queue."""
-        try:
-            self.return_path
-        except:
-            self.initMessage()
         if Defaults.PENDING_DELETE_APPEND:
-            Util.append_to_file(self.return_path,
+            Util.append_to_file(self.append_address,
                                 Defaults.PENDING_DELETE_APPEND)
         os.unlink(self.msgfile)
 
     def whitelist(self):
         """Whitelist the message sender."""
-        try:
-            self.return_path
-        except:
-            self.initMessage()
         if Defaults.PENDING_WHITELIST_APPEND:
-            Util.append_to_file(self.return_path,
+            Util.append_to_file(self.append_address,
                                 Defaults.PENDING_WHITELIST_APPEND)
         else:
             raise Errors.ConfigError, \
@@ -491,12 +476,8 @@ class Message:
 
     def blacklist(self):
         """Blacklist the message sender."""
-        try:
-            self.return_path
-        except:
-            self.initMessage()
         if Defaults.PENDING_BLACKLIST_APPEND:
-            Util.append_to_file(self.return_path,
+            Util.append_to_file(self.append_address,
                                 Defaults.PENDING_BLACKLIST_APPEND)
         else:
             raise Errors.ConfigError, \
@@ -508,7 +489,12 @@ class Message:
 
     def show(self):
         """Return the string representation of a message."""
-        return Util.msg_as_string(self.msgobj)
+        try:
+            return Util.msg_as_string(self.msgobj)
+        except TypeError:
+            # Re-parse using HeaderParser if Generator fails.
+            self.msgobj = Util.msg_from_file(open(self.msgfile, 'r'))
+            return Util.msg_as_string(self.msgobj)
 
     def getDate(self):
         timestamp = self.msgid.split('.')[0]
@@ -557,7 +543,7 @@ class Message:
     def summary(self, count = 0, total = 0, mailto = 0):
         """Return summary header information."""
         if not self.msg_size:
-            self.msg_size = len((Util.msg_as_string(self.msgobj)))
+            self.msg_size = len(self.show())
             if  self.msg_size == 1:
                 self.bytes =    self.bytes[:-1]
         str = self.msgid + " ("
