@@ -1,6 +1,6 @@
 # -*- python -*-
 #
-# Copyright (C) 2001,2002 Jason R. Mastaler <jason@mastaler.com>
+# Copyright (C) 2001,2002,2003 Jason R. Mastaler <jason@mastaler.com>
 #
 # This file is part of TMDA.
 #
@@ -22,7 +22,11 @@
 """General purpose functions."""
 
 
+from cStringIO import StringIO
+from string import whitespace as WHITESPACE
 import cPickle
+import email
+import email.Utils
 import fileinput
 import fnmatch
 import os
@@ -31,7 +35,6 @@ import pwd
 import re
 import socket
 import stat
-import string
 import sys
 import tempfile
 import time
@@ -39,14 +42,18 @@ import time
 import Errors
 
 
-EMPTYSTRING = ''
-NL = '\n'
 DOT = '.'
-from string import whitespace as WHITESPACE
+EMPTYSTRING = ''
+MODE_EXEC = 01
+MODE_READ = 04
+MODE_WRITE = 02
+NL = '\n'
 
 
 def gethostname():
-    hostname = os.environ.get('QMAILHOST') or \
+    """The host name"""
+    hostname = os.environ.get('TMDAHOST') or \
+               os.environ.get('QMAILHOST') or \
                os.environ.get('MAILHOST')
     if not hostname:
         hostname = socket.getfqdn()
@@ -54,18 +61,22 @@ def gethostname():
 
 
 def getfullname():
-    fullname = os.environ.get('QMAILNAME') or \
+    """The user's personal name.  Default is an empty value."""
+    fullname = os.environ.get('TMDANAME') or \
+               os.environ.get('QMAILNAME') or \
                os.environ.get('NAME') or \
                os.environ.get('MAILNAME')
     if not fullname:
         fullname = pwd.getpwuid(os.getuid())[4]
-    if not fullname:
-        fullname = ''
+        if fullname:
+            fullname = fullname.split(',')[0]
     return fullname
 
 
 def getusername():
-    username = os.environ.get('QMAILUSER') or \
+    """The user name"""
+    username = os.environ.get('TMDAUSER') or \
+               os.environ.get('QMAILUSER') or \
                os.environ.get('USER') or \
                os.environ.get('LOGNAME')
     if not username:
@@ -116,6 +127,85 @@ def getfileuid(path):
     return statinfo[stat.ST_UID]
 
 
+def getvdomainprepend(address, vdomainsfile):
+    ret_prepend = ''
+    if os.path.exists(vdomainsfile):
+        fp = open(vdomainsfile, 'r')
+        # Parse the virtualdomains control file; see qmail-send(8) for
+        # syntax rules.  All this because qmail doesn't store the original
+        # envelope recipient in the environment.
+        u, d = address.split('@', 1)
+        ousername = u.lower()
+        odomain = d.lower()
+        for line in fp.readlines():
+            vdomain_match = 0
+            line = line.strip().lower()
+            # Comment or blank line?
+            if line == '' or line[0] in '#':
+                continue
+            vdomain, prepend = line.split(':', 1)
+            # domain:prepend
+            if vdomain == odomain:
+                vdomain_match = 1
+            # .domain:prepend (wildcard)
+            elif vdomain[:1] == '.' and odomain.find(vdomain) != -1:
+                vdomain_match = 1
+            # user@domain:prepend
+            else:
+                try:
+                    if vdomain.split('@', 1)[1] == odomain:
+                        vdomain_match = 1
+                except IndexError:
+                    pass
+            if vdomain_match:
+                ret_prepend = prepend
+                break
+        fp.close()
+    return ret_prepend
+
+
+def getvuserhomedir(user, domain, script):
+    """Return the home directory of a qmail virtual domain user."""
+    cmd = "%s %s %s" % (script, user, domain)
+    fpin = os.popen(cmd)
+    vuserhomedir = fpin.read()
+    fpin.close()
+    return vuserhomedir.strip()
+
+
+def getuserparams(login):
+    "Return a user's home directory, UID, & GID."
+
+    stats = pwd.getpwnam(login)
+    return (stats[5], stats[2], stats[3])
+
+
+def RunTask(Args):
+    """Run a program the "hard way" so we don't lose our UID."""
+
+    # Open a pipe between the parent and a child process
+    Read, Write = os.pipe()
+    if not os.fork():
+        # Child writes only and can close the reader
+        os.close(Read)
+
+        # Capture the STDOUT and stick it in the pipe
+        os.dup2(Write, 1)
+
+        # Launch the program
+        os.execv(Args[0], Args)
+
+    # Parent reads only and can close the writer
+    os.close(Write)
+
+    # Capture contents of pipe
+    Read = os.fdopen(Read)
+    RetVal = Read.readlines()
+    Read.close()
+
+    return RetVal
+
+
 def seconds(timeout):
     """Translate the defined timeout interval into seconds."""
     match = re.match("^([0-9]+)([YMwdhms])$", timeout)
@@ -175,15 +265,15 @@ def unixdate(timesecs=None):
         timesecs = time.time()
     timetuple = time.localtime(timesecs)
     tzname = time.tzname[timetuple[-1]]
-    asctime_list = string.split(time.asctime(timetuple))
+    asctime_list = time.asctime(timetuple).split()
     asctime_list.insert(len(asctime_list)-1, tzname)
-    return string.join(asctime_list)
+    return ' '.join(asctime_list)
 
 
 def make_msgid(timesecs=None, pid=None):
     """Return an rfc2822 compliant Message-ID: string, composed of
     seconds since the epoch in UTC + process id + 'TMDA' + FQDN. e.g:
-    
+
     <1016659379.10104.TMDA@nightshade.la.mastaler.com>
 
     timesecs is optional, and if not given, the current time is used.
@@ -195,93 +285,39 @@ def make_msgid(timesecs=None, pid=None):
     if not pid:
         import Defaults
         pid = Defaults.PID
-    idhost = os.environ.get('QMAILIDHOST') or gethostname()
-    msgid = '<%s.%s.TMDA@%s>' % (int(timesecs), pid, idhost)
+    idhost = os.environ.get('TMDAIDHOST') or \
+             os.environ.get('QMAILIDHOST') or \
+             gethostname()
+    idtag = os.environ.get('TMDAIDTAG') or 'TMDA'
+    msgid = '<%s.%s.%s@%s>' % (int(timesecs), pid, idtag, idhost)
     return msgid
 
 
 def make_date(timesecs=None):
     """Return an RFC 2822 compliant Date: string.  e.g,
 
-    Thu, 16 May 2002 04:23:10 +1200 (NZST)
-    
+    Thu, 24 Apr 2003 10:15:15 +1200 (NZST)
+
     timesecs is optional, and if not given, the current time is used.
-
-    JRM: Once the email module is included within TMDA, we can use
-    utilize email.Utils.formatdate.
-    
-    Based on code from Python's email package
-    <URL:http://www.python.org/doc/current/lib/module-email.html>
-    Copyright (C) 2001,2002 Python Software Foundation.
     """
-    if not timesecs:
+    if timesecs is None:
         timesecs = time.time()
-    import Defaults
-    localtime = Defaults.LOCALDATE
-    if localtime:
-        now = time.localtime(timesecs)
-        # Calculate timezone offset, based on whether the local zone has
-        # daylight savings time, and whether DST is in effect.
-        if time.daylight and now[-1]:
-            offset = time.altzone
-        else:
-            offset = time.timezone
-        hours, minutes = divmod(abs(offset), 3600)
-        # Remember offset is in seconds west of UTC, but the timezone is in
-        # minutes east of UTC, so the signs differ.
-        if offset > 0:
-            sign = '-'
-        else:
-            sign = '+'
-        zone = '%s%02d%02d' % (sign, hours, minutes / 60)
-        tzname = time.tzname[now[-1]]
-    else:
-        now = time.gmtime(timesecs)
-        # Timezone offset is always -0000
-        zone = '-0000'
-        tzname = 'UTC'
-    return '%s, %02d %s %04d %02d:%02d:%02d %s (%s)' % (
-        ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun'][now[6]],
-        now[2],
-        ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
-         'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'][now[1] - 1],
-        now[0], now[3], now[4], now[5], zone, tzname)
-
-
-def formataddr(pair):
-    """The inverse of parseaddr(), this takes a 2-tuple of the form
-    (realname, email_address) and returns the string value suitable
-    for an RFC 2822 From:, To: or Cc:.
-    
-    If the first element of pair is false, then the second element is
-    returned unmodified.
-
-    JRM: Once the email mod is included within TMDA, we can nuke this
-    function and use email.Utils.formataddr.
-    """
-    specialsre = re.compile(r'[][\()<>@,:;".]')
-    escapesre = re.compile(r'[][\()"]')
-    name, address = pair
-    if name:
-        quotes = ''
-        if specialsre.search(name):
-            quotes = '"'
-        name = escapesre.sub(r'\\\g<0>', name)
-        return '%s%s%s <%s>' % (quotes, name, quotes, address)
-    return address
+    tzname = time.tzname[time.localtime(timesecs)[-1]]
+    basedate = email.Utils.formatdate(timesecs, localtime=1)
+    return '%s (%s)' % (basedate, tzname)
 
 
 def file_to_dict(file, dict):
     """Process and add then each line of a textfile to a dictionary."""
     for line in fileinput.input(file):
-        line = string.strip(line)
+        line = line.strip()
         # Comment or blank line?
         if line == '' or line[0] in '#':
             continue
         else:
-            fields = string.split(line)
+            fields = line.split()
             key = fields[0]
-            key = string.lower(key)
+            key = key.lower()
             value = fields[1]
             dict[key] = value
     return dict
@@ -291,14 +327,12 @@ def file_to_list(file):
     """Process and then append each line of file to list."""
     list = []
     for line in fileinput.input(file):
-        line = string.strip(line)
+        line = line.strip()
         # Comment or blank line?
         if line == '' or line[0] in '#':
             continue
         else:
-            line = string.expandtabs(line)
-            line = string.split(line, ' #')[0]
-            line = string.strip(line)
+            line = line.expandtabs().split('#')[0].strip()
             list.append(line)
     return list
 
@@ -311,7 +345,7 @@ def pipecmd(command, *strings):
     *strings are optional pieces of data to write to command.
 
     return_status will just return the exit status of the command..
-    
+
     Based on code from getmail
     <URL:http://www.qcc.sk.ca/~charlesc/software/getmail-2.0/>
     Copyright (C) 2001 Charles Cazabon, and licensed under the GNU
@@ -369,22 +403,24 @@ def writefile(contents, fullpathname):
         file = open(fullpathname, 'w')
         file.write(contents)
         file.close()
-        
+
 
 def append_to_file(str, fullpathname):
     """Append a string to a text file if it isn't already in there."""
     if os.path.exists(fullpathname):
         for line in fileinput.input(fullpathname):
-            line = string.lower(string.strip(line))
+            line = line.strip().lower()
             # Comment or blank line?
             if line == '' or line[0] in '#':
                 continue
             else:
-                if string.lower(string.strip(str)) == line:
+                line = line.expandtabs().split('#')[0].strip()
+                bare = str.expandtabs().split('#')[0].strip()
+                if bare.lower() == line:
                     fileinput.close()
                     return 0
     file = open(fullpathname, 'a+')
-    file.write(string.strip(str) + '\n')
+    file.write(str.strip() + '\n')
     file.close()
 
 
@@ -405,32 +441,236 @@ def pager(file):
     os.spawnvp(os.P_WAIT, pager_list[0], pager_list)
 
 
-def sendmail(headers, body, envrecip, envsender):
+def normalize_sender(sender):
+    """Return a normalized version of the given sender address for use
+    in ~/.tmda/responses.
+
+    - Any / characters are replaced with : to prevent creation of files
+      outside the directory.
+    - Spaces are replaced with underscores.
+    - The address is lowercased.
+    """
+    sender = sender.replace(' ', '_')
+    sender = sender.replace('/', ':')
+    return sender.lower()
+
+
+def confirm_append_address(xp, rp):
+    """
+    xp is an address from the ``X-Primary-Address'' header.
+    rp is the envelope sender address.
+
+    Compare the two addresses, and return the address appropriate for
+    CONFIRM_APPEND use based on the PRIMARY_ADDRESS_MATCH setting.
+    """
+    if not xp:
+        return rp
+    import Defaults
+    rpl = rp.lower()
+    xpl = xp.lower()
+    rplocal, rphost = rpl.split('@', 1)
+    rpdomain = '.'.join(rphost.split('.')[-2:])
+    rpusername = rplocal.split(Defaults.RECIPIENT_DELIMITER)[0]
+    xplocal, xphost = xpl.split('@', 1)
+    xpdomain = '.'.join(xphost.split('.')[-2:])
+    xpusername = xplocal.split(Defaults.RECIPIENT_DELIMITER)[0]
+    match = Defaults.PRIMARY_ADDRESS_MATCH
+    if match == 0:
+        # never a match
+        return rp
+    elif match == 1:
+        # only identical addresses match
+        if xpl == rpl:
+            return xp
+    elif match == 2:
+        # usernames and hostnames must match
+        if xpusername == rpusername and xphost == rphost:
+            return xp
+    elif match == 3:
+        # usernames and domains must match
+        if xpusername == rpusername and xpdomain == rpdomain:
+            return xp
+    elif match == 4:
+        # hostnames must match
+        if xphost == rphost:
+            return xp
+    elif match == 5:
+        # domains must match
+        if xpdomain == rpdomain:
+            return xp
+    elif match == 6:
+        # always a match
+        return xp
+    return rp
+
+
+def msg_from_file(fp, strict=0):
+    """Read a file and parse its contents into a Message object model.
+    Replacement for email.message_from_file().
+    
+    We use the HeaderParser subclass instead of Parser to avoid trying
+    to parse the message body, instead setting the payload to the raw
+    body as a string.  This is faster, and also helps us avoid
+    problems trying to parse spam with broken MIME bodies."""
+    from email.Message import Message
+    from email.Parser import HeaderParser
+    msg = HeaderParser(Message, strict=strict).parse(fp)
+    msg.header_parsed = 1
+    return msg
+
+
+def msg_as_string(msg, maxheaderlen=0, mangle_from_=0, unixfrom=0):
+    """A more flexible replacement for Message.as_string().  The default
+    is a textual representation of the message where the headers are
+    not wrapped, From is not escaped, and a leading From_ line is not
+    added.
+
+    msg is an email.Message.Message object.
+
+    maxheaderlen specifies the longest length for a non-continued
+    header.  Disabled by default.  RFC 2822 recommends 78.
+    
+    mangle_from_ escapes any line in the body that begins with "From"
+    with ">".  Useful when writing to Unix mbox files.  Default is
+    False.
+
+    unixfrom forces the printing of the envelope header delimiter.
+    Default is False."""
+    from email import Generator
+    fp = StringIO()
+    if hasattr(msg, 'header_parsed') and msg.header_parsed:
+        genclass = Generator.HeaderParsedGenerator
+    else:
+        genclass = Generator.Generator
+    g = genclass(fp, mangle_from_=mangle_from_, maxheaderlen=maxheaderlen)
+    g.flatten(msg, unixfrom=unixfrom)
+    return fp.getvalue()
+
+
+def sendmail(msgstr, envrecip, envsender):
     """Send e-mail via direct SMTP, or by opening a pipe to the
     sendmail program.
-    
-    headers can be either an rfc822.Message instance, or a set of
-    rfc822 compatible message headers as a string.
 
-    body is the message body content as a string.
+    msgstr is an rfc2822 message as a string.
 
     envrecip is the envelope recipient address.
 
     envsender is the envelope sender address.
     """
     import Defaults
-    if Defaults.OUTGOINGMAIL == 'smtp':
+    if Defaults.MAIL_TRANSPORT == 'sendmail':
+        # You can avoid the shell by passing a tuple of arguments as
+        # the command instead of a string.  This will cause the
+        # popen2.Popen3() code to execvp() "/usr/bin/sendmail" with
+        # these arguments exactly, with no trip through any shell.
+        cmd = (Defaults.SENDMAIL_PROGRAM, '-i',
+               '-f', envsender, '--', envrecip)
+        pipecmd(cmd, msgstr)
+    elif Defaults.MAIL_TRANSPORT == 'smtp':
         import SMTP
         server = SMTP.Connection()
-        server.sendmail(envsender, envrecip, (str(headers) + '\n' + body))
+        server.sendmail(envsender, envrecip, msgstr)
         server.quit()
-    elif Defaults.OUTGOINGMAIL == 'sendmail':
-        cmd = "%s -f '%s' -- '%s'" % (Defaults.SENDMAIL_PROGRAM,
-                                      envsender, envrecip)
-        pipecmd(cmd, str(headers), '\n', body)
     else:
         raise Errors.ConfigError, \
-              "Invalid OUTGOINGMAIL method: " + Defaults.OUTGOINGMAIL
+              "Invalid MAIL_TRANSPORT method: " + Defaults.MAIL_TRANSPORT
+
+
+def decode_header(str):
+    """Accept a possibly encoded message header as a string, and
+    return a decoded string if it can be decoded.
+
+    JRM: email.Header has a decode_header method, but it returns a
+    list of decoded pairs, one for each part of the header, which is
+    an awkward interface IMO, especially when the header contains a
+    mix of encoded and non-encoded parts.
+    """
+    try:
+        from email import Header
+        parts = []
+        pairs = Header.decode_header(str)
+        for pair in pairs:
+            parts.append(pair[0])
+        decoded_string = ' '.join(parts)
+        return decoded_string
+    except email.Errors.HeaderParseError:
+        return str
+
+
+def headers_as_list(msg):
+    """Return a list containing the entire set of header lines, in the
+    order in which they were read."""
+    return ['%s: %s' % (k, v) for k, v in msg.items()]
+
+
+def headers_as_raw_string(msg):
+    """Return the headers as a raw (undecoded) string."""
+    msgtext = msg_as_string(msg)
+    idx = msgtext.index('\n\n')
+    return msgtext[:idx+1]
+
+
+def headers_as_string(msg):
+    """Return the (decoded) message headers as a string.  If the
+    sequence can't be decoded, punt and return a raw (undecoded)
+    string instead."""
+    try:
+        hdrstr = '\n'.join(['%s: %s' %
+                            (k, decode_header(v)) for k, v in msg.items()])
+    except email.Errors.HeaderParseError:
+        hdrstr = headers_as_raw_string(msg)
+    return hdrstr
+
+
+def body_as_raw_string(msg):
+    """Return the body as a raw (undecoded) string."""
+    msgtext = msg_as_string(msg)
+    idx = msgtext.index('\n\n')
+    return msgtext[idx+2:]
+
+
+
+def rename_headers(msg, old, new):
+    """Rename all occurances of a message header in a Message object.
+
+    msg is an email.Message.Message object.
+
+    old is name of the header to rename.
+
+    new is the new name of the header
+    """
+    if msg.has_key(old):
+        for pair in msg._headers:
+            if pair[0].lower() == old.lower():
+                index = msg._headers.index(pair)
+                msg._headers[index] = (new, '%s' % pair[1])
+
+
+def add_headers(msg, headers):
+    """Add headers to a Message object.
+
+       msg is an email.Message.Message object.
+
+       headers is a dictionary of headers and values.
+       """
+    if headers:
+        keys = headers.keys()
+        keys.sort()
+        for k in keys:
+            del msg[k]
+            msg[k] = headers[k]
+
+
+def purge_headers(msg, headers):
+    """Purge headers from a Message object.
+
+       msg is an email.Message.Message object.
+
+       headers is a list of headers.
+       """
+    if headers:
+        for h in headers:
+            del msg[h]
 
 
 def build_cdb(filename):
@@ -438,7 +678,9 @@ def build_cdb(filename):
     import cdb
     try:
         cdbname = filename + '.cdb'
-        cdb = cdb.cdbmake(cdbname, cdbname + '.tmp')
+        tempfile.tempdir = os.path.dirname(filename)
+        tmpname = os.path.split(tempfile.mktemp())[1]
+        cdb = cdb.cdbmake(cdbname, cdbname + '.' + tmpname)
         for line in file_to_list(filename):
             linef = line.split()
             key = linef[0].lower()
@@ -455,16 +697,15 @@ def build_cdb(filename):
 
 
 def build_dbm(filename):
+    """Build a DBM file from a text file."""
     import anydbm
     import glob
-
     try:
-        (dbmpath, dbmname) = os.path.split(filename)
+        dbmpath, dbmname = os.path.split(filename)
         dbmname += '.db'
         tempfile.tempdir = dbmpath
         tmpname = tempfile.mktemp()
-
-        dbm = anydbm.open(tmpname,'n')
+        dbm = anydbm.open(tmpname, 'n')
         for line in file_to_list(filename):
             linef = line.split()
             key = linef[0].lower()
@@ -474,10 +715,9 @@ def build_dbm(filename):
                 value = ''
             dbm[key] = value
         dbm.close()
-
         for f in glob.glob(tmpname + '*'):
             (tmppath, tmpname) = os.path.split(tmpname)
-            newf = f.replace(tmpname,dbmname)
+            newf = f.replace(tmpname, dbmname)
             newf = os.path.join(tmppath, newf)
             os.rename(f, newf)
     except:
@@ -506,6 +746,21 @@ def unpickle(file):
     return object
 
 
+def db_insert(db, insert_sql, params):
+    """Insert (using the 'insert_sql' SQL) an address into a SQL DB."""
+    dbmodule = sys.modules[db.__module__]
+    DatabaseError = getattr(dbmodule, 'DatabaseError')
+    cursor = db.cursor()
+    try:
+        try:
+            cursor.execute(insert_sql, params)
+            db.commit()
+        except DatabaseError:
+            pass
+    finally:
+        cursor.close()
+
+
 def findmatch(list, addrs):
     """Determine whether any of the passed e-mail addresses match a
     Unix shell-style wildcard pattern contained in list.  The
@@ -513,13 +768,13 @@ def findmatch(list, addrs):
     the string if it exists (for exp and ext addresses only)."""
     for address in addrs:
         if address:
-            address = string.lower(address)
+            address = address.lower()
             for p in list:
-                stringparts = string.split(p)
+                stringparts = p.split()
                 p = stringparts[0]
                 # Handle special @=domain.dom syntax.
                 try:
-                    at = string.rindex(p, '@')
+                    at = p.rindex('@')
                     atequals = p[at+1] == '='
                 except (ValueError, IndexError):
                     atequals = None
@@ -619,16 +874,24 @@ def wraptext(text, column=70, honor_leading_ws=1):
 def maketext(templatefile, vardict):
     """Make some text from a template file.
 
-     Several locations are scanned for templatefile, in the following order:
-     1. The directory specified by tmda-filter's `-t' option.
-     2. Defaults.TEMPLATE_DIR
-     3. ../templates/
-     4. The package/RPM template directory.
+    templatefile can either be an absolute pathname starting with an /
+    or ~ (e.g, /usr/local/packages/tmda/templates/bounce.txt) or a
+    relative pathname (e.g, bounce.txt).
+
+    Given a relative pathname, several locations are scanned for
+    templatefile, in the following order:
+
+    1. The directory specified by tmda-filter's `-t' option.
+    2. Defaults.TEMPLATE_DIR_MATCH_SENDER (if true)
+    3. Defaults.TEMPLATE_DIR_MATCH_RECIPIENT (if true)
+    4. Defaults.TEMPLATE_DIR
+    5. ../templates/
+    6. The package/RPM template directories.
 
     The first match found stops the search.  In this way, you can
     specialize templates at the desired level, or, if you use only the
     default templates, you don't need to change anything.
-    
+
     Once the templatefile is found, string substitution is performed
     by interpolation in `localdict'.
 
@@ -638,21 +901,57 @@ def maketext(templatefile, vardict):
     and licensed under the GNU General Public License version 2.
     """
     import Defaults
-    # Calculate the locations to scan.
-    searchdirs = []
-    searchdirs.append(os.environ.get('TMDA_TEMPLATE_DIR'))
-    searchdirs.append(Defaults.TEMPLATE_DIR)
-    searchdirs.append(Defaults.PARENTDIR + '/templates/')
-    searchdirs.append(sys.prefix + '/share/tmda/')
-    searchdirs.append('/etc/tmda/')
-    # Start scanning.
     foundit = None
-    for dir in searchdirs:
-        if dir:
-            filename = os.path.join(dir, templatefile)
-            if os.path.exists(filename):
-                foundit = filename
-                break
+    if (templatefile[0] == '/' or templatefile[0] == '~'):
+        if templatefile[0] == '~':
+            templatefile = os.path.expanduser(templatefile)
+        if os.path.exists(templatefile):
+            foundit = templatefile
+    else:
+        # Calculate the locations to scan.
+        searchdirs = []
+        searchdirs.append(os.environ.get('TMDA_TEMPLATE_DIR'))
+        if Defaults.TEMPLATE_DIR_MATCH_SENDER and Defaults.TEMPLATE_DIR:
+            sender = os.environ.get('SENDER').lower()
+            searchdirs.append(os.path.join(Defaults.TEMPLATE_DIR, sender))
+            try:
+                domainparts = sender.split('@', 1)[1].split('.')
+                for i in range(len(domainparts)):
+                    searchdirs.append(os.path.join
+                                      (Defaults.TEMPLATE_DIR, '.'.join(domainparts)))
+                    del domainparts[0]
+            except IndexError:
+                pass
+        if Defaults.TEMPLATE_DIR_MATCH_RECIPIENT and Defaults.TEMPLATE_DIR:
+            recipient = os.environ.get('TMDA_RECIPIENT').lower()
+            searchdirs.append(os.path.join(Defaults.TEMPLATE_DIR, recipient))
+            try:
+                recippart, domainpart = recipient.split('@',1)
+                recipparts = recippart.split(Defaults.RECIPIENT_DELIMITER)
+                for i in range(len(recipparts)):
+                    searchdirs.append(os.path.join
+                                      (Defaults.TEMPLATE_DIR,
+                                       Defaults.RECIPIENT_DELIMITER.join(recipparts) +
+                                       "@" + domainpart))
+                    del recipparts[-1]
+                domainparts = domainpart.split('.')
+                for i in range(len(domainparts)):
+                    searchdirs.append(os.path.join
+                                      (Defaults.TEMPLATE_DIR, '.'.join(domainparts)))
+                    del domainparts[0]
+            except IndexError:
+                pass
+        searchdirs.append(Defaults.TEMPLATE_DIR)
+        searchdirs.append(os.path.join(Defaults.PARENTDIR, 'templates'))
+        searchdirs.append(os.path.join(sys.prefix, 'share/tmda'))
+        searchdirs.append('/etc/tmda/')
+        # Start scanning.
+        for dir in searchdirs:
+            if dir:
+                filename = os.path.join(dir, templatefile)
+                if os.path.exists(filename):
+                    foundit = filename
+                    break
     if foundit is None:
         raise IOError, "Can't find " + templatefile
     else:
@@ -663,12 +962,13 @@ def maketext(templatefile, vardict):
         localdict.update(vardict)
         text = template % localdict
         return text
-        
+
 
 def filter_match(filename, recip, sender=None):
     """Check if the give e-mail addresses match lines in filename."""
-    import FilterParser 
-    filter = FilterParser.FilterParser()
+    import Defaults
+    import FilterParser
+    filter = FilterParser.FilterParser(Defaults.DB_CONNECTION)
     filter.read(filename)
     (actions, matchline) = filter.firstmatch(recip, [sender])
     # print the results
@@ -684,3 +984,97 @@ def filter_match(filename, recip, sender=None):
         print 'MATCH:', matchline
     else:
         print 'Sorry, no matching lines.'
+
+
+def CanRead( file, uid = None, gid = None, raiseError = 1 ):
+    try:
+        return CanMode( file, MODE_READ, uid, gid )
+    except IOError:
+        if not raiseError:
+            return 0
+        else:
+            pass
+
+
+def CanWrite( file, uid = None, gid = None, raiseError = 1 ):
+    try:
+        return CanMode( file, MODE_WRITE, uid, gid )
+    except IOError:
+        if not raiseError:
+            return 0
+        else:
+            pass
+
+
+def CanExec( file, uid = None, gid = None, raiseError = 1 ):
+    try:
+        return CanMode( file, MODE_EXEC, uid, gid )
+    except IOError:
+        if not raiseError:
+            return 0
+        else:
+            pass
+
+
+def CanMode( file, mode = MODE_READ, uid = None, gid = None ):
+    try:
+        fstat = os.stat( file )
+    except:
+        raise IOError, "'%s' does not exist" % file
+    if uid is None:
+        uid = os.geteuid()
+    if gid is None:
+        gid = os.getegid()
+    needuid = fstat[4]
+    needgid = fstat[5]
+    filemod = fstat[0] & 0777
+    if uid == 0:
+        # Root always wins.
+        return 1
+    elif filemod & mode:
+        return 1
+    elif filemod & ( mode * 010 ) and needgid == gid:
+        return 1
+    elif filemod & ( mode * 0100 ) and needuid == uid:
+        return 1
+    else:
+        return 0
+
+
+class DevnullOutput:
+    def write(self, msg): pass
+    def flush(self): pass
+    def __repr__(self):
+        return ""
+
+
+class StringOutput:
+    def __init__(self):
+        self.__content = ""
+    def write(self, msg):
+        if msg != "":
+            self.__content += "%s" % msg
+    def flush(self):
+        self.__content = ""
+    def __repr__(self):
+        return self.__content
+
+
+class Debugable:
+    def __init__(self, outputObject = DevnullOutput() ):
+        self.DEBUGSTREAM = outputObject
+        if self.DEBUGSTREAM is DevnullOutput:
+            self.level = 0
+        else:
+            self.level = 1
+
+    def debug(self, msg, level = 1):
+        if self.level >= level:
+            print >> self.DEBUGSTREAM, msg
+
+    def set_debug(level = 1):
+        self.level = level
+
+    def set_nodebug():
+        self.level = 0
+
